/* Fiasco - computes sample statistics.
   Copyright (C) 1997 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The author may be contacted at <pfaffben@pilot.msu.edu> on the
   Internet, or as Ben Pfaff, 12167 Airport Rd, DeWitt MI 48820, USA
   through more mundane means. */

#include <config.h>
#include <assert.h>
#include <stdlib.h>
#include "common.h"
#include "error.h"
#include "file-handle.h"
#include "misc.h"
#include "settings.h"
#include "sfm.h"
#include "tokens.h"
#include "var.h"

#undef DEBUGGING
/*#define DEBUGGING 1*/
#include "debug-print.h"

/* XSAVE transformation (and related SAVE procedure). */
typedef struct
  {
    trns_header h;
    file_handle *f;		/* Associated system file. */
    int nvar;			/* Number of variables. */
    int *var;			/* Indices of variables. */
    flt64 *case_buf;		/* Case transfer buffer. */
  }
save_trns;

/* Options bits set by trim_dictionary(). */
#define GTSV_OPT_COMPRESSED	001	/* Compression; (X)SAVE only. */
#define GTSV_OPT_SAVE		002	/* The (X)SAVE procedure. */
#define GTSV_NONE		0

/* The file being read by the input program. */
static file_handle *get_file;

/* The transformation being used by the SAVE procedure. */
static save_trns *trns;

static int trim_dictionary (dictionary * dict, int *options);
static void read_from_get (void);
static int save_write_case_func (ccase *);
static int save_trns_proc (any_trns *, ccase *);
static void save_trns_free (any_trns *);

#if DEBUGGING
void dump_dict_variables (dictionary *);
#endif

/* Parses the GET command. */
int
cmd_get (void)
{
  file_handle *handle;
  dictionary *dict;
  int options = GTSV_NONE;

  int i;
  int nval;

  match_id (GET);
  discard_variables ();

  if (match_id (FILE))
    match_tok ('=');

  handle = fh_parse_file_handle ();
  if (handle == NULL)
    return 0;

  dict = sfm_read_dictionary (handle, NULL);
  if (dict == NULL)
    return 0;

#if DEBUGGING
  dump_dict_variables (dict);
#endif
  if (0 == trim_dictionary (dict, &options))
    {
      fh_close_handle (handle);
      return 0;
    }
#if DEBUGGING
  dump_dict_variables (dict);
#endif

  /* Set the fv and lv elements of all variables remaining in the
     dictionary. */
  nval = 0;
  for (i = 0; i < dict->nvar; i++)
    {
      variable *v = dict->var[i];

      v->fv = nval;
      nval += v->nv;
    }
  dict->nval = nval;
  assert (nval);

#if DEBUGGING
  printf (_("GET translation table from file to memory:\n"));
  for (i = 0; i < dict->nvar; i++)
    {
      variable *v = dict->var[i];

      printf (_("  %8s from %3d,%3d to %3d,%3d\n"), v->name,
	      v->p.get.fv, v->p.get.nv, v->fv, v->nv);
    }
#endif

  restore_dictionary (dict);

  vfm_source = &get_source;
  get_file = handle;

  return 1;
}

/* Parses the SAVE (for X==0) and XSAVE (for X==1) commands.  */
/* FIXME: save_dictionary() is too expensive.  It would make more
   sense to copy just the first few fields of each variables (up to
   `prv_index'): that's a SMOP. */
int
cmd_save_internal (int x)
{
  file_handle *handle;
  dictionary *dict;
  int options = GTSV_OPT_SAVE;

  save_trns *t;
  sfm_write_info inf;

  int i;

  match_id (SAVE);

  if (match_id (OUTFILE))
    match_tok ('=');

  handle = fh_parse_file_handle ();
  if (handle == NULL)
    return 0;

  dict = save_dictionary ();
#if DEBUGGING
  dump_dict_variables (dict);
#endif
  for (i = 0; i < dict->nvar; i++)
    dict->var[i]->prv_index = i;
  if (0 == trim_dictionary (dict, &options))
    {
      fh_close_handle (handle);
      return 0;
    }

#if DEBUGGING
  dump_dict_variables (dict);
#endif

  /* Write dictionary. */
  inf.h = handle;
  inf.pri = dict;
  inf.sec = &default_dict;
  inf.compress = !!(options & GTSV_OPT_COMPRESSED);
  if (!sfm_write_dictionary (&inf))
    {
      free_dictionary (dict);
      fh_close_handle (handle);
      return 0;
    }

  /* Fill in transformation structure. */
  t = trns = xmalloc (sizeof (save_trns));
  t->h.proc = save_trns_proc;
  t->h.free = save_trns_free;
  t->f = handle;
  t->nvar = dict->nvar;
  t->var = xmalloc (sizeof (int) * dict->nvar);
  for (i = 0; i < dict->nvar; i++)
    t->var[i] = dict->var[i]->prv_index;
  t->case_buf = xmalloc (sizeof (flt64) * inf.case_size);
  free_dictionary (dict);

  if (x == 0)
    /* SAVE. */
    {
      procedure (NULL, save_write_case_func, NULL);
      save_trns_free ((any_trns *) t);
    }
  else
    /* XSAVE. */
    add_transformation ((any_trns *) t);

  return 1;
}

/* Parses the SAVE procedure. */
int
cmd_save (void)
{
  return cmd_save_internal (0);
}

/* Parses the XSAVE transformation command. */
int
cmd_xsave (void)
{
  return cmd_save_internal (1);
}

static int
save_write_case_func (ccase * c)
{
  save_trns_proc ((any_trns *) trns, c);
  return 1;
}

static int
save_trns_proc (any_trns * t, ccase * c)
{
  flt64 *p = trns->case_buf;
  int i;

  for (i = 0; i < trns->nvar; i++)
    {
      variable *v = default_dict.var[trns->var[i]];
      if (v->type == NUMERIC)
	{
	  double src = c->data[v->fv].f;
	  if (src == SYSMIS)
	    *p++ = -FLT64_MAX;
	  else
	    *p++ = src;
	}
      else
	{
	  /* PORTME: See portability note 1. */
	  memcpy (p, c->data[v->fv].s, v->width);
	  memset (&((char *) p)[v->width], ' ',
		  REM_RND_UP (v->width, sizeof (flt64)));
	  p += DIV_RND_UP (v->width, sizeof (flt64));
	}
    }

  sfm_write_case (trns->f, trns->case_buf, p - trns->case_buf);
  return -1;
}

static void
save_trns_free (any_trns * trns)
{
  save_trns *t = (save_trns *) trns;

  fh_close_handle (t->f);
  free (t->var);
  free (t->case_buf);
  free (t);
}

/* Deletes NV variables from DICT, starting at index FIRST.  The
   variables must have consecutive indices.  The variables are cleared
   and freed. */
static void
dict_delete_run (dictionary *dict, int first, int nv)
{
  int i;

  for (i = first; i < dict->nvar - nv; i++)
    {
      dict->var[i] = dict->var[i + nv];
      dict->var[i]->index -= nv;
      clear_variable (dict, dict->var[i]);
      free (dict->var[i]);
    }
  dict->nvar -= nv;
}

static int rename_variables (dictionary * dict);

/* The GET and SAVE commands have a common structure after the
   FILE/OUTFILE subcommand.  This function parses this structure and
   returns nonzero on success, zero on failure.  It both reads
   *OPTIONS, for the GTSV_OPT_SAVE bit, and writes it, for the
   GTSV_OPT_COMPRESSED bit. */
/* FIXME: MAP. */
static int
trim_dictionary (dictionary *dict, int *options)
{
  *options = 0;
  if (set_scompression)
    *options |= GTSV_OPT_COMPRESSED;

  if (*options & GTSV_OPT_SAVE)
    {
      int i;
      
      /* Delete all the scratch variables. */
      for (i = 0; i < dict->nvar; i++)
	{
	  int j;
	  
	  if (dict->var[i]->name[0] != '#')
	    continue;

	  /* Find a run of variables to be deleted. */
	  for (j = i + 1; j < dict->nvar; j++)
	    if (dict->var[i]->name[0] != '#')
	      break;
	  
	  /* Actually delete 'em. */
	  dict_delete_run (dict, i, j - i);
	}
    }
  
  while (match_tok ('/'))
    {
      if (match_id (COMPRESSED))
	*options |= GTSV_OPT_COMPRESSED;
      else if (match_id (UNCOMPRESSED))
	*options &= ~GTSV_OPT_COMPRESSED;
      else if (match_id (DROP))
	{
	  variable **v;
	  int nv;
	  int i;

	  match_tok ('=');
	  if (!parse_variables (dict, &v, &nv, PV_NONE))
	    return 0;

	  /* Loop through the variables to delete. */
	  for (i = 0; i < nv;)
	    {
	      int j;

	      /* Find a run of variables to be deleted. */
	      for (j = i + 1; j < nv; j++)
		if (v[j]->index != v[j - 1]->index + 1)
		  break;

	      /* Actually delete 'em. */
	      dict_delete_run (dict, v[i]->index, j - i);
	      i = j;
	    }
	}
      else if (match_id (KEEP))
	{
	  variable **v;
	  int nv;
	  int i, i1, i2;

	  match_tok ('=');
	  if (!parse_variables (dict, &v, &nv, PV_NONE))
	    return 0;

	  /* Reorder the dictionary so that the kept variables are at
	     the beginning. */
	  for (i1 = 0; i1 < nv; i1++)
	    {
	      int i2 = v[i1]->index;

	      /* Swap variables with indices i1 and i2. */
	      variable *t = dict->var[i1];
	      dict->var[i1] = dict->var[i2];
	      dict->var[i2] = t;
	      dict->var[i1]->index = i1;
	      dict->var[i2]->index = i2;
	    }

	  /* Delete all but the first NV variables from the
	     dictionary. */
	  for (i = nv; i < dict->nvar; i++)
	    {
	      clear_variable (dict, dict->var[i]);
	      free (dict->var[i]);
	    }
	  dict->var = xrealloc (dict->var, sizeof (variable *) * nv);
	  dict->nvar = nv;
	}
      else if (match_id (RENAME))
	{
	  if (!rename_variables (dict))
	    return 0;
	}
      else
	return syntax_error (_("while expecting a valid subcommand"));

      if (dict->nvar == 0)
	{
	  msg (SE, _("All variables deleted from system file dictionary."));
	  return 0;
	}
    }

  if (token != '.')
    return msg (SE, _("Command terminator expected."));
  return 1;
}

/* Parses and performs the RENAME subcommand of GET and SAVE. */
static int
rename_variables (dictionary * dict)
{
  int i;

  int success = 0;

  variable **v;
  char **new_names;
  int nv, nn;

  int group;

  match_tok ('=');
  if (token != '(')
    {
      variable *v;

      v = parse_dict_variable (dict);
      if (v == NULL)
	return 0;
      force_match ('=');
      force_id ();
      if (!strncmp (tokstr, v->name, 8))
	return 1;
      if (is_dict_varname (dict, tokstr))
	return msg (SE, _("Cannot rename %s as %s because there already exists "
		    "a variable named %s.  To rename variables with "
		    "overlapping names, use a single RENAME subcommand such as "
		    "\"/RENAME (A=B)(B=C)(C=A)\", or equivalently, \"/RENAME "
		    "(A B C=B C A)\"."), v->name, tokstr, tokstr);
      rename_variable (dict, v, tokstr);
      get_token ();
      return 1;
    }

  nv = nn = 0;
  v = NULL;
  new_names = 0;
  group = 1;
  while (match_tok ('('))
    {
      int old_nv = nv;

      if (!parse_variables (dict, &v, &nv, PV_NO_DUPLICATE | PV_APPEND))
	goto lossage;
      if (!match_tok ('='))
	{
	  msg (SE, _("`=' expected after variable list."));
	  goto lossage;
	}
      if (!parse_DATA_LIST_vars (&new_names, &nn, PV_APPEND))
	goto lossage;
      if (nn != nv)
	{
	  msg (SE, _("Number of variables on left side of `=' (%d) do not "
	       "match number of variables on right side (%d), in "
	       "parenthesized group %d of RENAME subcommand."),
	       nv - old_nv, nn - old_nv, group);
	  goto lossage;
	}
      force_match (')');
      group++;
    }

  for (i = 0; i < nv; i++)
    force_avl_delete (dict->var_by_name, v[i]);
  for (i = 0; i < nv; i++)
    {
      strcpy (v[i]->name, new_names[i]);
      if (NULL != avl_insert (dict->var_by_name, v[i]))
	{
	  msg (SE, _("Duplicate variables name %s."), v[i]->name);
	  goto lossage;
	}
    }
  success = 1;

lossage:
  /* The label is a bit of a misnomer, we actually come here on any
     sort of return. */
  for (i = 0; i < nn; i++)
    free (new_names[i]);
  free (new_names);
  free (v);

  return success;
}

#if DEBUGGING
void
dump_dict_variables (dictionary * dict)
{
  int i;

  printf (_("\nVariables in dictionary:\n"));
  for (i = 0; i < dict->nvar; i++)
    printf ("%s, ", dict->var[i]->name);
  printf ("\n");
}
#endif

/* Clears internal state related to GET input procedure. */
static void
get_source_destroy_source (void)
{
  /* It is not necessary to destroy the dictionary because if we get
     to this point then the dictionary is default_dict. */
  fh_close_handle (get_file);
}

/* Reads all the cases from the data file and passes them to
   write_case(). */
static void
get_source_read (void)
{
  while (sfm_read_case (get_file, temp_case->data, &default_dict)
	 && write_case ())
    ;
  get_source_destroy_source ();
}

case_stream get_source =
  {
    NULL,
    get_source_read,
    NULL,
    NULL,
    get_source_destroy_source,
    NULL,
    "GET"
  };
