/* q2c - parser generator for Fiasco procedures.
   Copyright (C) 1997 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The author may be contacted at <pfaffben@pilot.msu.edu> on the
   Internet, or as Ben Pfaff, 12167 Airport Rd, DeWitt MI 48820, USA
   through more mundane means. */

/* Converts a `.q' file to a `.c' file.

   FIXME: Eventually this whole thing should be reimplemented as a
   flex/bison lexer/parser.  Right now it's full of feeping creaturism
   and code creep. */

#include <config.h>
#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>
#include <errno.h>
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#include "str.h"

/* Brokenness. */
#ifndef EXIT_SUCCESS
#define EXIT_SUCCESS 0
#endif

#ifndef EXIT_FAILURE
#define EXIT_FAILURE 1
#endif

#ifndef RAND_MAX
#undef rand
#undef srand
#define rand ansi_rand
#define srand ansi_srand
#define RAND_MAX 32767
#define BROKEN_RAND 1
#endif

#if !HAVE_STRERROR
#define strerror(X) 				\
	(sys_errlist[X])
#endif
     
/* I have decided that internationalization is too much of a hassle
   for the purposes of this mini-program. */
#undef _
#undef N_
#define _(x) x
#define N_(x) x

#undef DEBUGGING
/*#define DEBUGGING 1 */
#include "debug-print.h"

#define INBUFSZ 1024
#define TABSZ 256
#define TOKSZ 1024

char *pgmname;

int is_open;
FILE *in, *out;
char *ifn, *ofn;
int ln, oln = 1;
int bare;

char *buf, *cp;

#define STRING 256
#define ID     257
#define NUM    258		/* this is not used */
int token;
int tokint;
char *tokstr;

void
finish_up (void)
{
  if (!is_open)
    return;
  is_open = 0;
  fclose (in);
  fclose (out);
  if (remove (ofn) == -1)
    fprintf (stderr, _("%s: %s: remove: %s\n"), pgmname, ofn, strerror (errno));
}

int
fail (char *format,...)
{
  va_list args;

  va_start (args, format);
  fprintf (stderr, "%s: ", pgmname);
  vfprintf (stderr, format, args);
  fprintf (stderr, "\n");
  va_end (args);

  finish_up ();

  exit (EXIT_FAILURE);
}

int
error (char *format,...)
{
  va_list args;

  va_start (args, format);
  fprintf (stderr, _("%s:%d: (column %d) "), ifn, ln, (int) (cp - buf));
  vfprintf (stderr, format, args);
  fprintf (stderr, "\n");
  va_end (args);

  finish_up ();

  exit (EXIT_FAILURE);
}

/* halt-catch-fire */
void
hcf (void)
{
  finish_up ();
  exit (EXIT_FAILURE);
}

#define VME N_("virtual memory exhausted")

void *
xmalloc (size_t size)
{
  void *vp;
  if (size == 0)
    return NULL;
  vp = malloc (size);
  if (!vp)
    {
#if DEBUGGING && __CHECKER__
      error (_("xmalloc(%lu): Inducing segfault."), (unsigned long) size);
      *((int *) 0) = 0;
#endif
      fail ("xmalloc(%lu): %s", (unsigned long) size, VME);
    }
  return vp;
}

void *
xrealloc (void *ptr, size_t size)
{
  void *vp;
  if (!size)
    {
      if (ptr)
	free (ptr);
      return NULL;
    }
  if (ptr)
    vp = realloc (ptr, size);
  else
    vp = malloc (size);
  if (!vp)
    fail ("xrealloc(%lu): %s", (unsigned long) size, VME);
  return vp;
}

char *
xstrdup (char *s)
{
  size_t size;
  char *t;

  assert (s != NULL);
  size = strlen (s) + 1;
  t = malloc (size);
  if (!t)
    fail ("xstrdup(%lu): %s", (unsigned long) strlen (s), VME);
  memcpy (t, s, size);
  return t;
}

char *
strlower (char *s)
{
  static char *p;
  char *cp;
  if (!p)
    p = xmalloc (TOKSZ);
  cp = p;
  while (*s)
    *cp++ = tolower (*s++);
  *cp++ = 0;
  return p;
}

char *
strupper (char *s)
{
  static char *p;
  char *cp;
  if (!p)
    p = xmalloc (TOKSZ);
  cp = p;
  while (*s)
    *cp++ = toupper (*s++);
  *cp++ = 0;
  return p;
}

int
get_line (void)
{
  int len;

  ln++;
  if (0 == fgets (buf, INBUFSZ, in))
    {
      if (ferror (in))
	fail ("%s: fgets: %s", ifn, strerror (errno));
      return 0;
    }
  if (!strncmp ("/* *INDENT", buf, 10))
    {
      oln++;
      buf[0] = 0;
    }
  else if (buf[0] == '!')
    {
      buf[0] = 0;
      fputs (&buf[1], out);
      oln++;
    }
  else if (buf[0] == '$')
    buf[0] = ' ';
  len = strlen (buf);
  while (len && isspace (buf[len - 1]))
    buf[--len] = 0;
  cp = buf;
  return 1;
}

typedef struct symbol
{
  char *name;
  int unique;
  int ln;
  int value;
  struct symbol *next;
}
symbol;

symbol *symtab;

int
add_symbol (char *s, int unique, int value)
{
  symbol *iter, *sym;
  int x;

  sym = xmalloc (sizeof (symbol));
  sym->name = xstrdup (s);
  sym->unique = unique;
  sym->value = value;
  sym->next = NULL;
  sym->ln = ln;
  if (!symtab)
    {
      symtab = sym;
      return 1;
    }
  iter = symtab;
  x = 1;
  while (1)
    {
      if (streq (iter->name, s))
	{
	  if (iter->unique)
	    {
	      fprintf (stderr, _("%s:%d: `%s' is already defined above\n"), ifn,
		       ln, s);
	      fprintf (stderr, _("%s:%d: location of previous definition\n"), ifn,
		       iter->ln);
	      hcf ();
	    }
	  return x;
	}
      if (!iter->next)
	break;
      iter = iter->next;
      x++;
    }
  iter->next = sym;
  return ++x;
}

symbol *
find_symbol (int x)
{
  int y = x;
  symbol *iter;

  iter = symtab;
  while (x > 1 && iter)
    {
      iter = iter->next;
      x--;
    }
  if (!iter)
    fail (_("find_symbol(%d): internal error"), y);
  return iter;
}

#if DEBUGGING
int
dump_token (void)
{
  switch (token)
    {
    case NUM:
      printf ("NUM\t%d\n", tokint);
      break;
    case STRING:
      printf ("STRING\t\"%s\"\n", tokstr);
      break;
    case ID:
      printf ("ID\t%s\n", tokstr);
      break;
    default:
      printf ("PUNCT\t%c\n", token);
    }
}
#endif /* DEBUGGING */

int
get_token (void)
{
  while (1)
    {
      while (isspace (*cp))
	cp++;
      if (!*cp)
	{
	  if (!get_line ())
	    fail (_("%s: unexpected end of file"), ifn);
	}
      else
	break;
    }
  if (*cp == '_' || isalnum (*cp))
    {
      char *dest = tokstr;
      token = ID;
      while (*cp == '_' || isalnum (*cp))
	*dest++ = toupper (*cp++);	/* yeah, i regret that too */
      *dest++ = 0;
    }
  else if (*cp == '"')
    {
      char *dest = tokstr;
      token = STRING;
      cp++;
      while (*cp != '"' && *cp)
	{
	  if (*cp == '\\')
	    {
	      *dest++ = *cp++;
	      if (!*cp)
		error (_("unterminated string literal"));
	      *dest++ = *cp++;
	    }
	  else
	    *dest++ = *cp++;
	}
      *dest++ = 0;
      if (!*cp)
	error (_("unterminated string literal"));
      cp++;
    }
  /*
     else if(isdigit(*cp))
     {
     token = NUM;
     tokint = *cp++ - '0';
     while(isdigit(*cp))
     tokint = tokint * 10 + *cp++ - '0';
     }
   */
  else
    token = *cp++;
#if DEBUGGING
  dump_token ();
#endif
  return token;
}

void
force_id (void)
{
  if (token != ID)
    error (_("identifier expected"));
}

void
force_string (void)
{
  if (token != STRING)
    error (_("string expected"));
}

int
match_id (char *s)
{
  if (token == ID && streq (tokstr, s))
    {
      get_token ();
      return 1;
    }
  return 0;
}

int
match_token (int t)
{
  if (token == t)
    {
      get_token ();
      return 1;
    }
  return 0;
}

void
skip_token (int t)
{
  if (token != t)
    error (_("`%c' expected"), t);
  get_token ();
}

typedef enum
{
  VAL_NONE,
  VAL_INT,
  VAL_DBL,
}
val_type;

enum
  {
    VT_PLAIN,
    VT_PAREN,
  };

typedef struct
  {
    char *specname;
    int valtype;
    val_type value;
    int optvalue;
    char *valname;
    char *restriction;
    int def;
    int con;
    int def2;
  }
setting;

typedef struct
  {
    char *varname;
    setting *s;
    int ns;
    int index;
  }
specifier;

typedef enum
  {
    SBC_PLAIN,
    SBC_VARLIST,
    SBC_VARLIST_ALL,
    SBC_INT,
    SBC_PINT,
    SBC_DBL,
    SBC_INT_LIST,
    SBC_DBL_LIST,
    SBC_CUSTOM,
    SBC_ARRAY,
    SBC_STRING,
    SBC_VAR,
  }
subcommand_type;

typedef struct
  {
    char *name;
    char *restriction;		/* for SBC_STRING */
    char *message;		/* for SBC_STRING */
    subcommand_type type;
    int once;
    specifier *spec;
    int nspec;
    int narray;
    char *prefix;
  }
subcommand;

char *cmdname;
int ntab;
int def = -1;			/* index of default subcommand */
subcommand *tab;

void
parse_setting (setting * s, subcommand * sbc, specifier * spec)
{
  /* `*' indicates a setting for which the keyword is optional.
   * example: the TO setting on the CASES subcommand of LIST
   */
  if (match_token ('*'))
    s->def2 = 1;
  else
    s->def2 = 0;

  /* `!' indicates the setting the specifier is initialized to.
   */
  if (match_token ('!'))
    s->def = 1;
  else
    s->def = 0;

  force_id ();
  s->specname = xstrdup (tokstr);

  get_token ();
  if (token != '/' && token != ';' && token != '.' && token != ',')
    {
      s->con = add_symbol (s->specname, 0, 0);
      if (token == '(')
	{
	  s->valtype = VT_PAREN;
	  get_token ();
	}
      else
	s->valtype = VT_PLAIN;
      if (match_token ('*'))
	s->optvalue = 1;
      else
	s->optvalue = 0;
      if (match_id ("N"))
	s->value = VAL_INT;
      else if (match_id ("D"))
	s->value = VAL_DBL;
      else
	error (_("`n' or `d' expected"));
      skip_token (':');
      force_id ();
      s->valname = xstrdup (tokstr);
      get_token ();
      if (token == ',')
	{
	  get_token ();
	  force_string ();
	  s->restriction = xstrdup (tokstr);
	  get_token ();
	}
      else
	s->restriction = NULL;
      if (s->valtype == VT_PAREN)
	skip_token (')');
    }
  else
    {
      s->con = add_symbol (s->specname, 0, 0);
      s->value = VAL_NONE;
    }
}

void
parse_specifier (specifier * spec, subcommand * sbc)
{
  int nset, mset;

  if (token == ID)
    {
      spec->varname = xstrdup (strlower (tokstr));
      get_token ();
    }
  else
    spec->varname = NULL;
  spec->ns = 0;
  if (token != ':')
    {
      spec->index = sbc->narray;
      if (sbc->type == SBC_ARRAY)
	if (token == '|')
	  token = ',';
	else
	  sbc->narray++;
      spec->s = NULL;
      return;
    }
  spec->index = 0;
  skip_token (':');

  nset = 0;
  mset = 16;
  spec->s = xmalloc (sizeof (setting) * mset);
  do
    {
      if (nset >= mset)
	{
	  mset += 16;
	  spec->s = xrealloc (spec->s, sizeof (setting) * mset);
	}
      parse_setting (&spec->s[nset++], sbc, spec);
      if (token != ',' && token != ';' && token != '.')
	skip_token ('/');
    }
  while (token != ',' && token != ';' && token != '.');
  spec->s = xrealloc (spec->s, sizeof (setting) * nset);
  spec->ns = nset;
}

void
parse_specifiers (subcommand * sbc)
{
  int nspec, mspec;

  nspec = 0;
  mspec = 16;
  sbc->spec = xmalloc (sizeof (specifier) * mspec);
  sbc->narray = 0;

  while (token != ';' && token != '.')
    {
      if (nspec >= mspec)
	{
	  mspec += 16;
	  sbc->spec = xrealloc (sbc->spec, sizeof (specifier) * mspec);
	}
      parse_specifier (&sbc->spec[nspec++], sbc);
      if (token != ';' && token != '.')
	skip_token (',');
    }
  sbc->spec = xrealloc (sbc->spec, sizeof (specifier) * nspec);
  sbc->nspec = nspec;
}

#define BASE_INDENT 2
#define INC_INDENT 2

int indent;

void
dump (char *format,...)
{
  va_list args;
  int i;

  oln++;
  va_start (args, format);
  for (i = 0; i < indent; i++)
    putc (' ', out);
  vfprintf (out, format, args);
  putc ('\n', out);
  va_end (args);
}

void
dump_specifier_vars (specifier * spec, char *s, int i, int j, subcommand * sbc)
{
  int k, def = 0;

  if (spec->varname)
    {
      for (k = 0; k < spec->ns; k++)
	if (spec->s[k].def)
	  {
	    if (def)
	      error (_("multiple defaults for specifier `%s'"), spec->varname);
	    def = spec->s[k].con;
	    break;
	  }

      dump ("%slong %s%s;", s, sbc->prefix, spec->varname);
    }
  for (k = 0; k < spec->ns; k++)
    if (spec->s[k].value != VAL_NONE)
      {
	char *typename;
	if (spec->s[k].value == VAL_INT)
	  typename = "long";
	else if (spec->s[k].value == VAL_DBL)
	  typename = "double";
	else
	  fail (_("dump_specifier(): internal error"));
	dump ("%s%s %s%s;", s, typename, sbc->prefix,
	      strlower (spec->s[k].valname));
      }
}

int
is_keyword (char *t)
{
  char *kw[] =
  {"AND", "OR", "NOT", "EQ", "GE", "GT", "LE", "LT",
   "NE", "ALL", "BY", "TO", "WITH", 0};
  char **cp;

  for (cp = kw; *cp; cp++)
    if (streq (t, *cp))
      return 1;
  return 0;
}

char *
make_identifier (char *name)
{
  static char temp[64];
  char *cp;

  for (cp = temp; *name; name++)
    if (isalpha (*name))
      *cp++ = tolower (*name);
    else
      *cp++ = '_';
  return temp;
}

void
dump_vars (int local)
{
  char *s;
  int i, j, k, f;
  symbol *sym;

  indent = local ? BASE_INDENT : 0;
  s = local ? "" : "static ";
  for (sym = symtab, f = k = 0; sym; sym = sym->next)
    if (!sym->unique && !is_keyword (sym->name))
      {
	if (!f)
	  {
	    dump ("enum %s_tag", make_identifier (cmdname));
	    indent += INC_INDENT;
	    dump ("{");
	    indent += INC_INDENT;
	    f = 1;
	  }

	if (k)
	  dump ("%s,", sym->name);
	else
	  {
	    k = 1;
	    dump ("%s = 1000,", sym->name);
	  }
      }
  if (f)
    {
      indent -= INC_INDENT;
      dump ("};");
      indent -= INC_INDENT;
      dump ("");
    }

  for (i = 0; i < ntab; i++)
    if (tab[i].type == SBC_ARRAY && tab[i].narray)
      {
	char *prefix;

	dump ("enum %s_tag", make_identifier (tab[i].name));
	indent += INC_INDENT;
	dump ("{");
	indent += INC_INDENT;

	prefix = xstrdup (strupper (tab[i].prefix));
	for (j = 0; j < tab[i].nspec; j++)
	  if (!tab[i].spec[j].ns)
	    dump ("%s%s = %d,", prefix, strupper (tab[i].spec[j].varname),
		  tab[i].spec[j].index);
	free (prefix);

	indent -= INC_INDENT;
	dump ("};");
	indent -= INC_INDENT;
	dump ("");
      }
  for (i = 0; i < ntab; i++)
    {
      int f = 0;

      dump (_("/* %s subcommand */"), tab[i].name);
      dump ("%sint sbc_%s;", s, strlower (tab[i].name));
      if (tab[i].type == SBC_ARRAY || tab[i].type == SBC_PLAIN)
	for (j = 0; j < tab[i].nspec; j++)
	  if (tab[i].spec[j].s == 0)
	    {
	      if (tab[i].type == SBC_PLAIN)
		dump ("%slong int %s%s;", s, strlower (tab[i].prefix),
		      tab[i].spec[j].varname);
	      else if (f == 0)
		{
		  dump ("%sint %sa_%s[%d];", s, tab[i].prefix,
			strlower (tab[i].name),
			tab[i].narray);
		  f = 1;
		}
	    }
	  else
	    dump_specifier_vars (&tab[i].spec[j], s, i, j, &tab[i]);
      else if (tab[i].type == SBC_VARLIST || tab[i].type == SBC_VARLIST_ALL)
	{
	  char *n = xstrdup (strlower (tab[i].name));
	  dump ("%sint %sn_%s;", s, strlower (tab[i].prefix), n);
	  dump ("%svariable **%sv_%s;", s, strlower (tab[i].prefix), n);
	  free (n);
	}
      else if (tab[i].type == SBC_VAR)
	{
	  char *n = xstrdup (strlower (tab[i].name));
	  dump ("%svariable *%sv_%s;", s, strlower (tab[i].prefix), n);
	}
      else if (tab[i].type == SBC_STRING)
	dump ("%schar *s_%s;", s, strlower (tab[i].name));
      else if (tab[i].type == SBC_INT || tab[i].type == SBC_PINT)
	dump ("%slong n_%s;", s, strlower (tab[i].name));
      dump ("");
    }
}

void
dump_specifier_init (specifier * spec, subcommand * sbc)
{
  int k, def = 0;

  if (spec->varname)
    {
      char *s;

      for (k = 0; k < spec->ns; k++)
	if (spec->s[k].def)
	  {
	    def = spec->s[k].con;
	    break;
	  }
      if (def)
	s = find_symbol (def)->name;
      else
	s = "-1";
      dump ("%s%s = %s;", sbc->prefix, spec->varname, s);
    }
  for (k = 0; k < spec->ns; k++)
    if (spec->s[k].value != VAL_NONE)
      {
	char *init;
	if (spec->s[k].value == VAL_INT)
	  init = "NOT_LONG";
	else if (spec->s[k].value == VAL_DBL)
	  init = "SYSMIS";
	else
	  fail (_("dump_specifier_init(): internal error"));
	dump ("%s%s = %s;", sbc->prefix, strlower (spec->s[k].valname), init);
      }
}

void
dump_vars_init (void)
{
  int i, j;

  for (i = 0; i < ntab; i++)
    for (j = 0; j < tab[i].nspec; j++)
      if (tab[i].spec[j].s == 0 && tab[i].type == SBC_ARRAY)
	{
	  dump ("int i;");
	  goto breakout;
	}
breakout:
  for (i = 0; i < ntab; i++)
    {
      int f = 0;

      dump ("sbc_%s = 0;", strlower (tab[i].name));
      if (tab[i].type == SBC_PLAIN || tab[i].type == SBC_ARRAY)
	for (j = 0; j < tab[i].nspec; j++)
	  if (tab[i].spec[j].s == 0)
	    {
	      if (tab[i].type == SBC_PLAIN)
		dump ("%s%s = 0;", tab[i].prefix, tab[i].spec[j].varname);
	      else if (f == 0)
		{
		  dump ("for(i=0; i<%d; i++)", tab[i].narray);
		  indent += INC_INDENT;
		  dump ("%sa_%s[i] = 0;", tab[i].prefix,
			strlower (tab[i].name));
		  indent -= INC_INDENT;
		  f = 1;
		}
	    }
	  else
	    dump_specifier_init (&tab[i].spec[j], &tab[i]);
      else if (tab[i].type == SBC_VARLIST || tab[i].type == SBC_VARLIST_ALL)
	{
	  char *p = xstrdup (strlower (tab[i].prefix));
	  dump ("%sn_%s = 0;", p, strlower (tab[i].name));
	  dump ("%sv_%s = NULL;", p, strlower (tab[i].name));
	  free (p);
	}
      else if (tab[i].type == SBC_VAR)
	{
	  char *p = xstrdup (strlower (tab[i].prefix));
	  dump ("%sv_%s = NULL;", p, strlower (tab[i].name));
	  free (p);
	}
      else if (tab[i].type == SBC_STRING)
	dump ("s_%s = NULL;", strlower (tab[i].name));
      else if (tab[i].type == SBC_INT || tab[i].type == SBC_PINT)
	dump ("n_%s = NOT_LONG;", strlower (tab[i].name));
    }
}

char *
make_match (char *t)
{
  static char *s;

  if (!s)
    s = xmalloc (TOKSZ);
  if (is_keyword (t))
    {
      sprintf (s, "match_tok (%s)", t);
      return s;
    }
  if (isdigit (t[0]))
    {
      sprintf (s, "match_int (%s)", t);
      return s;
    }
  if (streq (t, "ON") || streq (t, "YES"))
    strcpy (s, "(match_id (ON) || match_id (YES) || match_id (TRUE))");
  else if (streq (t, "OFF") || streq (t, "NO"))
    strcpy (s, "(match_id (OFF) || match_id (NO) || match_id (FALSE))");
  else
    {
      while (*t == '_')
	t++;
      if (isdigit (t[0]))
	sprintf (s, "match_int (%s)", t);
      else
	sprintf (s, "match_id (%s)", t);
    }
  return s;
}

void
dump_specifier_parse (specifier * spec, subcommand * sbc, int j)
{
  int k;

  for (k = 0; k < spec->ns; k++)
    {
      if (spec->s[k].def2)
	{
	  if (k + 1 < spec->ns)
	    error (_("default setting is not last setting in `%s' specifier"),
		   spec->varname);
	  if (j + 1 < sbc->nspec)
	    error (_("default setting is not in last specifier in `%s' "
		   "subcommand"), sbc->name);
	  if (j || k)
	    {
	      dump ("else");
	      indent += INC_INDENT;
	      dump ("{");
	      indent += INC_INDENT;
	    }
	  dump ("%s;", make_match (spec->s[k].specname));
	}
      else
	{
	  dump ("%sif (%s)", (j || k) ? "else " : "",
		make_match (spec->s[k].specname));
	  indent += INC_INDENT;
	}
      if (spec->s[k].value == VAL_NONE)
	dump ("%s%s = %s;", sbc->prefix, spec->varname,
	      find_symbol (spec->s[k].con)->name);
      else
	{
	  if (!spec->s[k].def2)
	    {
	      dump ("{");
	      indent += INC_INDENT;
	    }
	  if (spec->varname)
	    dump ("%s%s = %s;", sbc->prefix, spec->varname,
		  find_symbol (spec->s[k].con)->name);
	  if (spec->s[k].valtype == VT_PAREN)
	    {
	      if (spec->s[k].optvalue)
		{
		  dump ("if (token == '(')");
		  indent += INC_INDENT;
		  dump ("{");
		  indent += INC_INDENT;
		}
	      else
		{
		  dump ("if (token != '(')");
		  indent += INC_INDENT;
		  dump (_("return msg (SE, \"`(' expected after %s specifier of "
			  "%s subcommand.\");"),
			spec->s[k].specname, sbc->name);
		  indent -= INC_INDENT;
		}
	      dump ("get_token ();");
	    }

	  if (spec->s[k].value == VAL_INT)
	    {
	      dump ("if (token != NUM || tokint == NOT_LONG)");
	      indent += INC_INDENT;
	      dump (_("return msg (SE, \"%s specifier of %s subcommand "
		      "requires an integer argument.\");"),
		    spec->s[k].specname, sbc->name, cmdname);
	      indent -= INC_INDENT;
	      dump ("%s%s = tokint;", sbc->prefix,
		    strlower (spec->s[k].valname));
	    }
	  else
	    {
	      dump ("if (token != NUM)");
	      indent += INC_INDENT;
	      dump (_("return msg (SE, \"Number expected after %s specifier of "
		      "%s subcommand.\");"),
		    spec->s[k].specname, sbc->name);
	      indent -= INC_INDENT;
	      dump ("%s%s = tokval;", sbc->prefix,
		    strlower (spec->s[k].valname));
	    }
	  if (spec->s[k].restriction)
	    {
	      char *s, *s2;
	      s = xmalloc (TOKSZ);
	      s2 = xmalloc (TOKSZ);
	      sprintf (s2, "%s%s", sbc->prefix, strlower (spec->s[k].valname));
	      sprintf (s, spec->s[k].restriction, s2);
	      dump ("if (!(%s))", s);
	      free (s);
	      free (s2);
	      indent += INC_INDENT;
	      dump (_("return msg (SE, \"Bad argument for `%s' specifier of `%s' "
		      "subcommand.\");"),
		    spec->s[k].specname, sbc->name);
	      indent -= INC_INDENT;
	    }
	  dump ("get_token ();");
	  if (spec->s[k].valtype == VT_PAREN)
	    {
	      dump ("if (token != ')')");
	      indent += INC_INDENT;
	      dump (_("return msg (SE, \"`)' expected after argument for %s "
		      "specifier of %s.\");"),
		    spec->s[k].specname, sbc->name);
	      indent -= INC_INDENT;
	      dump ("get_token ();");
	      if (spec->s[k].optvalue)
		{
		  indent -= INC_INDENT;
		  dump ("}");
		  indent -= INC_INDENT;
		}
	    }
	  if (!spec->s[k].def2)
	    {
	      indent -= INC_INDENT;
	      dump ("}");
	    }
	}
#if 0
      else
      {
	dump ("/* not implemented */");
	dump (_("return msg(SE, \"`%s' specifier of `%s' subcommand of `%s' "
	      "not currently implemented.\");"),
	      spec->s[k].specname, sbc->name, cmdname);
      }
#endif
      if (spec->s[k].def2)
	{
	  indent -= INC_INDENT;
	  dump ("}");
	  indent -= INC_INDENT;
	}
      indent -= INC_INDENT;
    }
}

void
dump_subcommand (subcommand * sbc)
{
  if (sbc->type == SBC_PLAIN || sbc->type == SBC_ARRAY)
    {
      int j, n;

      dump ("while (token != '/' && token != '.')");
      indent += INC_INDENT;
      dump ("{");
      indent += INC_INDENT;

      for (j = n = 0; j < sbc->nspec; n += sbc->spec[j].ns, j++)
	if (sbc->spec[j].ns)
	  dump_specifier_parse (&sbc->spec[j], sbc, j);
	else
	  {
	    n++;
	    dump ("%sif (%s)", j ? "else " : "",
		  make_match (strupper (sbc->spec[j].varname)));
	    indent += INC_INDENT;
	    if (sbc->type == SBC_PLAIN)
	      dump ("%s%s = 1;", strlower (sbc->prefix), sbc->spec[j].varname);
	    else
	      {
		char *s = xstrdup (strupper (sbc->prefix));
		dump ("%sa_%s[%s%s] = 1;", sbc->prefix, strlower (sbc->name),
		      s, strupper (sbc->spec[j].varname));
		free (s);
	      }
	    indent -= INC_INDENT;
	  }

      /* FIXME: this is definitely not right.
         We need this to be generated in most circumstances.  Sometimes
         (if n==0) the `else' is not needed.
       */
      if (sbc->nspec
	  && (sbc->spec[sbc->nspec - 1].ns == 0
	      || (sbc->spec[sbc->nspec - 1].s
		  [sbc->spec[sbc->nspec - 1].ns - 1].def2 == 0)))
	dump (_("%s return syntax_error (\"in %s subcommand\");"),
	      n ? "else " : "", sbc->name);

      dump ("match_tok (',');");
      indent -= INC_INDENT;
      dump ("}");
      indent -= INC_INDENT;
    }
  else if (sbc->type == SBC_VARLIST || sbc->type == SBC_VARLIST_ALL)
    {
      char *s, *p;

      s = xstrdup (strlower (sbc->name));
      p = strlower (sbc->prefix);
      dump ("if (!parse_variables (NULL, &%sv_%s, &%sn_%s, PV_APPEND%s%s))",
	    p, s, p, s,
	    sbc->message ? " |" : "",
	    sbc->message ? sbc->message : "");
      free (s);
      indent += INC_INDENT;
      dump ("return 0;");
      indent -= INC_INDENT;
    }
  else if (sbc->type == SBC_VAR)
    {
      char *s, *p;

      s = xstrdup (strlower (sbc->name));
      p = strlower (sbc->prefix);
      dump ("%sv_%s = parse_variable ();", p, s);
      dump ("if (%sv_%s)", p, s);
      free (s);
      indent += INC_INDENT;
      dump ("return 0;");
      indent -= INC_INDENT;
    }
  else if (sbc->type == SBC_STRING)
    {
      if (sbc->restriction)
	{
	  dump ("{");
	  indent += INC_INDENT;
	  dump ("int x;");
	}
      dump ("force_string ();");
      if (sbc->restriction)
	{
	  dump ("x = strlen (tokstr);");
	  dump ("if (!(%s))", sbc->restriction);
	  indent += INC_INDENT;
	  dump (_("return msg (SE, \"String for %s must be %s.\");"),
		sbc->name, sbc->message);
	  indent -= INC_INDENT;
	}
      dump ("s_%s = xstrdup (tokstr);", strlower (sbc->name));
      dump ("get_token ();");
      if (sbc->restriction)
	{
	  indent -= INC_INDENT;
	  dump ("}");
	}
    }
  else if (sbc->type == SBC_INT)
    {
      dump ("force_int ();");
      dump ("n_%s = tokint;", strlower (sbc->name));
    }
  else if (sbc->type == SBC_PINT)
    {
      dump ("match_tok ('(');");
      dump ("force_int ();");
      dump ("n_%s = tokint;", strlower (sbc->name));
      dump ("match_tok (')');");
    }
  else if (sbc->type == SBC_CUSTOM)
    {
      dump ("switch (custom_%s ())", strlower (sbc->name));
      indent += INC_INDENT;
      dump ("{");
      dump ("case 0:");
      indent += INC_INDENT;
      dump ("return 0;");
      indent -= INC_INDENT;
      dump ("case 2:");
      indent += INC_INDENT;
      dump (_("return syntax_error (\"in %s subcommand\");"),
	    sbc->name);
      indent -= INC_INDENT;
      dump ("}");
      indent -= INC_INDENT;
    }
}

void
dump_parser (void)
{
  char *cp;
  int f, i;

  indent = BASE_INDENT;
  dump (_("/* parser for %s command begins */"), cmdname);
  if (!bare)
    {
      dump ("get_token ();");

      cp = strtok (cmdname, " ");
      cp = strtok (NULL, " ");
      while (cp)
	{
	  dump ("force_match_id (%s);", strupper (cp));
	  cp = strtok (NULL, " ");
	}
    }

  dump ("{");
  indent += INC_INDENT;

  dump_vars_init ();

  dump ("while(1)");
  indent += INC_INDENT;
  dump ("{");
  indent += INC_INDENT;

  f = 0;
  if (def != -1 && (tab[def].type == SBC_VARLIST
		    || tab[def].type == SBC_VARLIST_ALL))
    {
      char *p, *s;

      if (tab[def].type == SBC_VARLIST)
	dump ("if (token == ID && is_varname (tokstr) && lookahead () != '=')");
      else
	{
	  dump ("if ((token == ID && is_varname (tokstr) && lookahead () != '=')");
	  dump ("     || token == ALL)");
	}
      indent += INC_INDENT;
      dump ("{");
      indent += INC_INDENT;
      dump ("sbc_%s++;", strlower (tab[def].name));
      s = xstrdup (strlower (tab[def].name));
      p = strlower (tab[def].prefix);
      dump ("if (!parse_variables (NULL, &%sv_%s, &%sn_%s, PV_APPEND))",
	    p, s, p, s);
      free (s);
      indent += INC_INDENT;
      dump ("return 0;");
      indent -= INC_INDENT;
      indent -= INC_INDENT;
      dump ("}");
      indent -= INC_INDENT;
      f = 1;
    }
  else if (def != -1 && tab[def].type == SBC_CUSTOM)
    {
      dump ("switch (custom_%s ())", strlower (tab[def].name));
      indent += INC_INDENT;
      dump ("{");
      dump ("case 0:");
      indent += INC_INDENT;
      dump ("return 0;");
      indent -= INC_INDENT;
      dump ("case 1:");
      indent += INC_INDENT;
      dump ("sbc_%s++;", strlower (tab[def].name));
      dump ("continue;");
      indent -= INC_INDENT;
      dump ("}");
      indent -= INC_INDENT;
    }
  for (i = 0; i < ntab; i++)
    {
      dump ("%sif (%s)", f ? "else " : "", make_match (tab[i].name));
      f = 1;
      indent += INC_INDENT;
      dump ("{");

      indent += INC_INDENT;
      dump ("match_tok ('=');");
      dump ("sbc_%s++;", strlower (tab[i].name));
      if (tab[i].once)
	{
	  dump ("if (sbc_%s > 1)", strlower (tab[i].name));
	  indent += INC_INDENT;
	  dump (_("return msg (SE, \"%s subcommand may be given only once.\");"),
		tab[i].name);
	  indent -= INC_INDENT;
	}
      dump_subcommand (&tab[i]);
      indent -= INC_INDENT;
      dump ("}");
      indent -= INC_INDENT;
    }

  indent -= INC_INDENT;
  dump ("if (token == '/')");
  indent += INC_INDENT;
  dump ("get_token ();");
  indent -= INC_INDENT;
  dump ("else break;");

  indent -= INC_INDENT;
  dump ("}");
  indent -= INC_INDENT;
/*  if(!bare) */
  {
    dump ("if (token != '.')");
    indent += INC_INDENT;
    dump (_("return msg (SE, \"`/' or end of command expected.\");"));
    indent -= INC_INDENT;
  }
  indent -= INC_INDENT;
  dump ("}");
  dump (_("/* parser for %s command ends */"), cmdname);
}

void
dump_header (void)
{
  time_t curtime;
  struct tm *loctime;
  char *timep;

  indent = 0;
  curtime = time (NULL);
  loctime = localtime (&curtime);
  timep = asctime (loctime);
  timep[strlen (timep) - 1] = 0;
  dump ("/*");
  dump (" *  %s", ofn);
  dump (" *");
  dump (_(" *  Generated by q2c from %s on %s."), ifn, timep);
  dump (_(" *  Do not modify!"));
  dump (" */");
  dump ("");
}

void
dump_free (void)
{
  int i;

  indent = BASE_INDENT;
  for (i = 0; i < ntab; i++)
    if (tab[i].type == SBC_STRING)
      dump ("free (s_%s);", strlower (tab[i].name));
}

int
main (int argc, char *argv[])
{
  pgmname = argv[0];
  if (argc != 3)
    fail (_("syntax: q2c input.q output.c"));

  ifn = argv[1];
  in = fopen (ifn, "r");
  if (!in)
    fail ("%s: open: %s", ifn, strerror (errno));

  ofn = argv[2];
  out = fopen (ofn, "w");
  if (!out)
    fail ("%s: open: %s", ofn, strerror (errno));

  is_open = 1;
  buf = xmalloc (INBUFSZ);
  tokstr = xmalloc (TOKSZ);
  tab = xmalloc (sizeof (subcommand[TABSZ]));

  /* prime the lexer */
  get_line ();
  get_token ();

  /* get the command name */
  if (token != ID && token != STRING)
    error (_("command name expected"));
  cmdname = xstrdup (tokstr);
  get_token ();
  skip_token (':');
  while (token != '.')
    {
      subcommand *sbc = &tab[ntab];

      /* `*' means that that subcommand may be
       * given w/o its identifying keyword.
       */
      if (token == '*')
	{
	  if (def != -1)
	    error (_("multiple default subcommands"));
	  def = ntab;
	  get_token ();
	}
      /* `+' means a subcommand may be given only
       * once.
       */
      if (token == '+')
	{
	  sbc->once = 1;
	  get_token ();
	}
      else
	sbc->once = 0;

      force_id ();
      sbc->name = xstrdup (tokstr);
      sbc->narray = 0;
      sbc->nspec = 0;
      sbc->type = SBC_PLAIN;

      get_token ();
      if (token == '[')
	{
	  get_token ();
	  force_id ();
	  sbc->prefix = xstrdup (strlower (tokstr));
	  get_token ();
	  skip_token (']');
	  skip_token ('=');
	  sbc->type = SBC_ARRAY;
	  parse_specifiers (sbc);
	}
      else
	{
	  if (token == '(')
	    {
	      get_token ();
	      force_id ();
	      sbc->prefix = xstrdup (strlower (tokstr));
	      get_token ();
	      skip_token (')');
	    }
	  else
	    sbc->prefix = "";
	  skip_token ('=');

	  if (match_id ("VAR"))
	    sbc->type = SBC_VAR;
	  if (match_id ("VARLIST"))
	    {
	      if (match_token ('('))
		{
		  force_string ();
		  sbc->message = xstrdup (tokstr);
		  get_token();
		  skip_token (')');
		}
	      else sbc->message = NULL;
	      
	      if (match_id ("ALL"))
		sbc->type = SBC_VARLIST_ALL;
	      else
		sbc->type = SBC_VARLIST;
	    }
	  else if (match_id ("INTEGER"))
	    if (match_id ("LIST"))
	      sbc->type = SBC_INT_LIST;
	    else
	      sbc->type = SBC_INT;
	  else if (match_id ("PINT"))
	    sbc->type = SBC_PINT;
	  else if (match_id ("DOUBLE"))
	    if (match_id ("LIST"))
	      sbc->type = SBC_DBL_LIST;
	    else
	      sbc->type = SBC_DBL;
	  else if (match_id ("STRING"))
	    {
	      sbc->type = SBC_STRING;
	      if (token == STRING)
		{
		  sbc->restriction = xstrdup (tokstr);
		  get_token ();
		  force_string ();
		  sbc->message = xstrdup (tokstr);
		  get_token ();
		}
	      else
		sbc->restriction = NULL;
	    }
	  else if (match_id ("CUSTOM"))
	    sbc->type = SBC_CUSTOM;
	  else
	    parse_specifiers (sbc);
	}

      while (token == ';')
	get_token ();

      ntab++;
    }

  dump_header ();
  indent = 0;
  dump ("#line %d \"%s\"", ln + 1, ifn);
  while (get_line ())
    {
      char *cp = buf;

      while (isspace (*cp))
	cp++;
      if (!strncmp (cp, "/*(", 3))
	{
	  char *s;

	  s = strlower (cp);
	  cp = s + 3;
	  while (isspace (*cp))
	    cp++;
	  dump ("#line %d \"%s\"", oln + 1, ofn);
	  if (!strncmp (cp, "header", 6))
	    {
	      indent = 0;
	      dump ("#include \"str.h\"");
	      dump ("#include \"common.h\"");
	      dump ("#include \"error.h\"");
	      dump ("#include \"tokens.h\"");
	      dump ("#include \"var.h\"");
	      dump ("");
	    }
	  else if (!strncmp (cp, "decls", 5))
	    {
	      int local;
	      cp += 5;
	      while (isspace (*cp))
		cp++;
	      if (!strncmp (cp, "local", 5))
		local = 1;
	      else if (!strncmp (cp, "global", 6))
		local = 0;
	      else
		error (_("`local' or `global' expected"));
	      dump_vars (local);
	    }
	  else if (!strncmp (cp, "parser", 6))
	    {
	      cp += 6;
	      while (isspace (*cp))
		cp++;
	      if (!strncmp (cp, "bare", 4))
		bare = 1;
	      else
		bare = 0;
	      dump_parser ();
	    }
	  else if (!strncmp (cp, "free", 4))
	    dump_free ();
	  else
	    error (_("`decls', `header', `parser', or `free' expected"));
	  indent = 0;
	  dump ("#line %d \"%s\"", ln + 1, ifn);
	}
      else
	dump ("%s", buf);
    }

  return EXIT_SUCCESS;
}
