/* Fiasco - computes sample statistics.
   Copyright (C) 1997 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The author may be contacted at <pfaffben@pilot.msu.edu> on the
   Internet, or as Ben Pfaff, 12167 Airport Rd, DeWitt MI 48820, USA
   through more mundane means. */

#include <config.h>
#include <ctype.h>
#include <assert.h>
#include <stdarg.h>
#include <limits.h>
#include "error.h"
#include "str.h"
#include "misc.h"
#include "arena.h"
#include "output.h"
#include "som.h"
#include "var.h"

/* This file contains low level som table submission routines. */

/* Define virtual local variables. */
#define EXTERN extern
#include "somP.h"
#undef EXTERN

static void draw_table_cell (outp_driver * d, som_table * t,
			     int x, int y, int h, int v);

void som_get_table_size (outp_driver * d, som_table * t);
void som_dump_page (int x1, int y1, int x2, int y2, int px, int py, int trim);

/* Draws a single cell C on output device D in a box with upper-left
   corner (X,Y) and dimensions (H,V). */
static inline void
draw_cell (outp_driver * d, som_any_cell * c, int x, int y, int h, int v)
{
  switch (c->any.con)
    {
    case SCON_VALUE:
      {
	outp_text t;

	t.options = OUTP_T_HORZ | OUTP_T_VERT;
	switch (c->any.opt & SOPT_A_MASK)
	  {
	  case SOPT_A_RGHT:
	    t.options |= OUTP_T_JUST_RIGHT;
	    break;
	  case SOPT_A_LEFT:
	    t.options |= OUTP_T_JUST_LEFT;
	    break;
	  case SOPT_A_CNTR:
	    t.options |= OUTP_T_JUST_CENTER;
	    break;
	  case SOPT_A_JUST:
	    t.options |= OUTP_T_JUST_FULL;
	    break;
	  default:
	    assert (0);
	  }
	t.s.l = LOAD_2 (&c->text.s.l1);
	t.s.s = c->text.s.s;
	t.h = h;
	t.v = v;
	t.x = x;
	t.y = y;
	d->class->text_draw (d, &t);
      }
      break;
    case SCON_TEXT:
      {
	outp_text t;

	t.options = OUTP_T_HORZ | OUTP_T_VERT | OUTP_T_FANCY;
	switch (c->any.opt & SOPT_A_MASK)
	  {
	  case SOPT_A_RGHT:
	    t.options |= OUTP_T_JUST_RIGHT;
	    break;
	  case SOPT_A_LEFT:
	    t.options |= OUTP_T_JUST_LEFT;
	    break;
	  case SOPT_A_CNTR:
	    t.options |= OUTP_T_JUST_CENTER;
	    break;
	  case SOPT_A_JUST:
	    t.options |= OUTP_T_JUST_FULL;
	    break;
	  default:
	    assert (0);
	  }
	t.s.l = LOAD_2 (&c->text.s.l1);
	t.s.s = c->text.s.s;
	t.h = h;
	t.v = v;
	t.x = x;
	t.y = y;
	d->class->text_draw (d, &t);
      }
      break;
    case SCON_TABLE:
      draw_table_cell (d, c->table.t, x, y, h, v);
      break;
    case SCON_EMPTY:
      break;
    default:
      assert (0);
    }
}

/* Draws an intersection of two rules in a box with upper-left corner
   (X,Y) and dimensions (H,V); the intersection takes place in the
   table at location (C,R). */
static inline void
draw_intersection (som_table * t, int x, int y, int h, int v, int c, int r)
{
  color color =
  {0, 0, 0, 0};
  outp_styles styles;
  rect rect;

  if (h == 0 || v == 0)
    return;

  rect.x1 = x;
  rect.y1 = y;
  rect.x2 = x + h;
  rect.y2 = y + v;

  styles.t = (r > 0 ? t->rv[(c / 2) + (t->nc + 1) * (r / 2 - 1)]
	      : OUTP_L_NONE);
  styles.b = (r < 2 * t->nr ? t->rv[(c / 2) + (t->nc + 1) * (r / 2)]
	      : OUTP_L_NONE);
  styles.l = (c > 0 ? t->rh[(c / 2 - 1) + (t->nc) * (r / 2)]
	      : OUTP_L_NONE);
  styles.r = (c < 2 * t->nc ? t->rh[(c / 2) + (t->nc) * (r / 2)]
	      : OUTP_L_NONE);

  if (styles.t & SLIN_SPACING)
    styles.t = 0;
  if (styles.b & SLIN_SPACING)
    styles.b = 0;
  if (styles.l & SLIN_SPACING)
    styles.l = 0;
  if (styles.r & SLIN_SPACING)
    styles.r = 0;

  if (styles.t | styles.b | styles.l | styles.r)
    som.d->class->line_intersection (som.d, &rect, &color, &styles);
}

/* Draws a horizontal rule with line-style STYLE in a box with
   upper-left corner (X1,Y1) and dimensions (H,V). */
static inline void
draw_horz_rule (int x1, int y1, int h, int v, int style)
{
  color c =
  {0, 0, 0, 0};
  rect r;

  if (h == 0 || v == 0 || (style & SLIN_SPACING) || style == SLIN_0)
    return;
  r.x1 = x1;
  r.y1 = y1;
  r.x2 = x1 + h;
  r.y2 = y1 + v;
  som.d->class->line_horz (som.d, &r, &c, style);
}

/* Draws a vertical rule with line-style STYLE in a box with
   upper-left corner (X1,Y1) and dimensions (H,V). */
static inline void
draw_vert_rule (int x1, int y1, int h, int v, int style)
{
  color c =
  {0, 0, 0, 0};
  rect r;

  if (h == 0 || v == 0 || (style & SLIN_SPACING) || style == SLIN_0)
    return;
  r.x1 = x1;
  r.y1 = y1;
  r.x2 = x1 + h;
  r.y2 = y1 + v;
  som.d->class->line_vert (som.d, &r, &c, style);
}

/* Draws a cell that contains a table T on output device D in a box
   with upper left hand corner (H,K) and size (W,L). */
static void
draw_table_cell (outp_driver * d, som_table * t, int h, int k, int w, int l)
{
  int cp_x, cp_y;
  int x, y;

  cp_y = k;
  for (y = 0; y <= 2 * t->nr; y++)
    {
      cp_x = h;

      for (x = 0; x <= 2 * t->nc; x++)
	{
	  switch (ROW_TYPE (y) + 2 * COLUMN_TYPE (x))
	    {
	    case RULE_ROW + 2 * RULE_COLUMN:
	      draw_intersection (t, cp_x, cp_y, t->w[x], t->h[y], x, y);
	      break;
	    case CELL_ROW + 2 * RULE_COLUMN:
	      draw_vert_rule (cp_x, cp_y, t->w[x], t->h[y],
			      t->rv[x / 2 + (y / 2) * (t->nc + 1)]);
	      break;
	    case RULE_ROW + 2 * CELL_COLUMN:
	      draw_horz_rule (cp_x, cp_y, t->w[x], t->h[y],
			      t->rh[x / 2 + (y / 2) * t->nc]);
	      break;
	    case CELL_ROW + 2 * CELL_COLUMN:
	      draw_cell (som.d, t->c[x / 2 + t->nc * (y / 2)],
			 cp_x, cp_y, t->w[x], t->h[y]);
	      break;
	    }

	  cp_x += t->w[x];
	}

      cp_y += t->h[y];
    }
}

/* Draws a page of output.  The cells (X1,Y1)-(X2,Y2) inclusive are
   drawn.  Header cells are always drawn, with the exception that when
   bit 0 of TRIM is set, top header is not drawn, and when bit 1 is
   set, bottom header is not drawn.  Calls draw_title(PX,PY) to draw a
   title line if PX is not -2. */
void
som_dump_page (int x1, int y1, int x2, int y2, int px, int py, int trim)
{
  som_table *t = som.t;

  int i, j;
  int x, y;
  int c1, c2;
  int r1, r2;
  int i1, i2;
  int cp_x, cp_y;

  i1 = trim & 1;
  i2 = 3 - trim / 2;

  cp_x = som.ext->cp_x;
  cp_y = som.ext->cp_y;
  if (som.title_height)
    {
      if (px != -2)
	som_draw_title (px, py, 0);
      cp_y += som.title_height;
    }
  for (i = i1; i < i2; i++)
    {
      switch (i)
	{
	case 0:
	  r1 = 0;
	  r2 = 2 * t->t;
	  break;
	case 1:
	  r1 = y1;
	  r2 = y2;
	  break;
	case 2:
	  r1 = 2 * (t->nr - t->b);
	  r2 = 2 * t->nr;
	  break;
	}
      for (y = r1; y <= r2; y++)
	{
	  cp_x = som.ext->cp_x;

	  for (j = 0; j < 3; j++)
	    {
	      switch (j)
		{
		case 0:
		  c1 = 0;
		  c2 = 2 * t->l;
		  break;
		case 1:
		  c1 = x1;
		  c2 = x2;
		  break;
		case 2:
		  c1 = 2 * (t->nc - t->r);
		  c2 = 2 * t->nc;
		  break;
		}
	      for (x = c1; x <= c2; x++)
		{
		  switch (ROW_TYPE (y) + 2 * COLUMN_TYPE (x))
		    {
		    case RULE_ROW + 2 * RULE_COLUMN:
		      draw_intersection (t, cp_x, cp_y, t->w[x], t->h[y], x, y);
		      break;
		    case CELL_ROW + 2 * RULE_COLUMN:
		      draw_vert_rule (cp_x, cp_y, t->w[x], t->h[y],
				      t->rv[x / 2 + (y / 2) * (t->nc + 1)]);
		      break;
		    case RULE_ROW + 2 * CELL_COLUMN:
		      draw_horz_rule (cp_x, cp_y, t->w[x], t->h[y],
				      t->rh[x / 2 + (y / 2) * t->nc]);
		      break;
		    case CELL_ROW + 2 * CELL_COLUMN:
		      draw_cell (som.d, t->c[x / 2 + t->nc * (y / 2)],
				 cp_x, cp_y, t->w[x], t->h[y]);
		      break;
		    }

		  cp_x += t->w[x];
		}
	    }
	  cp_y += t->h[y];
	}
    }
  som.ext->cp_x = cp_x;
  som.ext->cp_y = cp_y;
}

/* Draws the subrow number SR_INDX at location (CP_X, CP_Y).  These
   subrow numbers help to make crushed tables with lots of subrows a
   little more readable. */
void
output_row_label (int cp_x, int cp_y, int sr_indx)
{
  outp_text text;
  char buf[INT_DIGITS + 2];

  /* Don't label the first row. */
  if (sr_indx == 1)
    return;

  sprintf (buf, "%d", sr_indx);
  text.options = OUTP_T_HORZ | OUTP_T_JUST_RIGHT;
  text.x = cp_x;
  text.y = cp_y;
  text.h = som.digit_space;
  text.s.l = strlen (buf);
  text.s.s = buf;
  som.d->class->text_draw (som.d, &text);
}

/* Outputs the crushed table from subrow SR1 up to but not including
   subrow SR2. */
void
som_dump_crush_page (int sr1, int sr2)
{
  som_table *t = som.t;
  som_driver_ext *ext = som.d->som;

  int sr;

  int cp_y = som.ext->cp_y;

  printf (_("Dumping subrows (%d,%d):\n"), sr1, sr2);
  
  for (sr = sr1; sr < sr2; sr++)
    {
      int row = SUBROW_TO_ROW (sr);
      int sr_indx = SUBROW_WITHIN_ROW (sr);
      int subrow_height = SUBROW_HEIGHT (sr);

      if (sr == sr1)
	printf ("%d", sr);
      else
	printf ("+%d", sr);
      if(sr == sr2 - 1)
	printf ("\n");
      
      /*
      printf ("som_dump_crush_page: row %d, subrow %d, subrow_height %d, cp_y=%d\n",
	      row, sr_indx, subrow_height, cp_y);
	      */
      
      assert (sr_indx >= 0 && sr_indx <= som.nch);
      if (sr_indx != 0)
	{
	  int column;
	  int cp_x = 0;
	  int rule_style, rule_width;

	  for (column = som.ch[sr_indx - 1]; column <= som.ch[sr_indx]; column++)
	    {
	      if (IS_CELL_COLUMN (column))
		draw_cell (som.d, t->c[column / 2 + t->nc * (row / 2)],
			   cp_x, cp_y, t->w[column], subrow_height);
	      else
		draw_vert_rule (cp_x, cp_y, t->w[column], subrow_height,
				t->rv[column / 2 + (row / 2) * (t->nc + 1)]);
	      cp_x += t->w[column];
	    }
	  if (som.nch > 1)
	    output_row_label (som.mpw, cp_y, sr_indx);

	  /* If the subrow is less wide than the row itself, draw a
	     vertical rule in the style of the rightmost rule in the
	     row, assuming there is room for it. */
	  rule_style = t->rv[t->nc + (row / 2) * (t->nc + 1)];
	  rule_width = t->w[2 * t->nc];
	  if (cp_x + rule_width < som.mpw)
	    draw_vert_rule (som.mpw - rule_width,
			    cp_y, rule_width, subrow_height, rule_style);
	}
      else
	{
	  int cp_x = 0;

	  draw_intersection (t, cp_x, cp_y, t->w[0], t->h[row], 0, row);
	  draw_horz_rule (cp_x + t->w[0], cp_y,
			  som.mpw - t->w[0] - t->w[t->nc * 2],
			  t->h[row], t->rh[(row / 2) * t->nc]);
	  draw_intersection (t, cp_x + som.mpw - t->w[t->nc * 2], cp_y,
			     t->w[t->nc * 2], t->h[row], t->nc * 2, row);
	}

      cp_y += subrow_height;
    }
  som.ext->cp_x = 0;
  som.ext->cp_y = cp_y;
}

/* Determines the size of cell C.  *CW is the cell width, *CH is the
   cell height.  If FIXED==0, both width and height are determined.
   If FIXED==1, width is calculated for fixed height, which must
   already be in *CH.  If FIXED==2, height is calculated for fixed
   width, which must already be in *CW. */
static void
get_cell_size (outp_driver * d, som_any_cell * c, int *cw, int *ch, int fixed)
{
  outp_text t;
  som_driver_ext *x = d->som;

  if (fixed != 2)
    {
      t.options = 0;
      switch (c->any.con)
	{
	case SCON_EMPTY:
	  *cw = 0;
	  break;
	case SCON_TEXT:
	  t.options = OUTP_T_FANCY;	/* fall through */
	case SCON_VALUE:
	  t.s.l = LOAD_2 (&c->text.s.l1);
	  t.s.s = c->text.s.s;
	  d->class->text_metrics (d, &t);
	  *cw = t.h;
	  break;
	case SCON_TABLE:
	  {
	    som_table *t = c->table.t;
	    int i;

	    som_get_table_size (d, t);
	    for (*cw = i = 0; i <= 2 * t->nc; i++)
	      *cw += t->w[i];
	  }
	  break;
	case SCON_JOIN_AUX:
	  assert (0);		/* should never occur */
	case SCON_GRID:
	  t.s.l = c->grid.nc;
	  t.s.s = c->grid.s;
	  d->class->text_metrics (d, &t);
	  *cw = t.h;
	  break;
	case SCON_DRAW:
	  assert (0);		/* not yet implemented */
	default:
	  assert (0);		/* internal error */
	}
    }
  if (fixed != 1)
    {
      switch (c->any.con)
	{
	case SCON_EMPTY:
	  *ch = 0;
	  break;
	case SCON_VALUE:
	  *ch = x->font_height;
	  break;
	case SCON_TEXT:
	  t.options = OUTP_T_FANCY | OUTP_T_HORZ;
	  t.s.l = LOAD_2 (&c->text.s.l1);
	  t.s.s = c->text.s.s;
	  t.h = *cw;
	  d->class->text_metrics (d, &t);
	  *ch = t.v;
	  break;
	case SCON_TABLE:
	  {
	    som_table *t = c->table.t;
	    int i;

	    if (fixed == 2)
	      som_get_table_size (d, t);
	    for (*ch = i = 0; i <= 2 * t->nr; i++)
	      *ch += t->h[i];
	  }
	  break;
	case SCON_JOIN_AUX:
	  assert (0);		/* should never occur */
	case SCON_GRID:
	case SCON_DRAW:
	  assert (0);		/* not yet implemented */
	default:
	  assert (0);		/* internal error */
	}
    }
}

#if GLOBAL_DEBUGGING
static void check_table (som_table * t);
#endif

/* Determines the size of each row and column in table T with respect
   to driver D. */
void
som_get_table_size (outp_driver * d, som_table * t)
{
  /* Driver extension. */
  som_driver_ext *som = d->som;

  /* 0-based coordinates of current cell in table. */
  int x, y;

  /* Row, column height/width vectors. */
  int *h, *w;
  int *hp, *wp;

  /* Current cell in table. */
  som_any_cell **c;

#if GLOBAL_DEBUGGING
  check_table (t);
#endif

  w = t->w;
  memset (w, 0, sizeof (int) * (2 * t->nc + 1));
  h = t->h;
  memset (h, 0, sizeof (int) * (2 * t->nr + 1));

  /* Check out the width/height of columns/rows proper.

     Deep ugliness begins here. */
  for (c = t->c, y = 0, hp = h + 1; y < t->nr; y++, hp += 2)
    for (x = 0, wp = w + 1; x < t->nc; x++, c++, wp += 2)
      if (((*c)->any.opt & SOPT_X_MASK) == SOPT_X_NORM)
	{
	  int cw, ch;
	  get_cell_size (d, *c, &cw, &ch, 0);
	  if (cw > *wp)
	    *wp = cw;
	  if (ch > *hp)
	    *hp = ch;
	}
      else if (((*c)->any.opt & SOPT_X_MASK) == SOPT_X_JOIN)
	{
	  /* add this cell to a linked list of joined cells?
	     perhaps just set a flag since joined cells are the exception? */
	}

  /* Check out the width/height of rules. */
  for (x = 0; x <= t->nc; x++)
    {
      int rw, mrw = 0;
      for (y = 0; y < t->nr; y++)
	{
	  rw = som->width[(t->rv[x + y * (t->nc + 1)]) & ~SLIN_SPACING];
	  if (rw > mrw)
	    mrw = rw;
	}
      t->w[2 * x] = mrw;
    }
  for (y = 0; y <= t->nr; y++)
    {
      int rh, mrh = 0;
      for (x = 0; x < t->nc; x++)
	{
	  rh = som->height[(t->rh[x + y * t->nc]) & ~SLIN_SPACING];
	  if (rh > mrh)
	    mrh = rh;
	}
      t->h[2 * y] = mrh;
    }

  for (c = t->c, y = 0, hp = h + 1; y < t->nr; y++, hp += 2)
    for (x = 0, wp = w + 1; x < t->nc; x++, c++, wp += 2)
      switch ((*c)->any.opt & SOPT_X_MASK)
	{
	case SOPT_X_NORM:
	case SOPT_X_NONE:
	  break;
	case SOPT_X_HORZ:
	  {
	    /* Changes horizontal size to whatever the cell wants.
	       Does not change vertical size.  */
	    int cw, ch;

	    ch = *hp;
	    get_cell_size (d, *c, &cw, &ch, 1);
	    if (cw > *wp)
	      *wp = cw;
	  }
	  break;
	case SOPT_X_VERT:
	  {
	    /* Changes vertical size to whatever the cell wants.  Does
	       not change horizontal size. */
	    int cw, ch;

	    cw = *wp;
	    get_cell_size (d, *c, &cw, &ch, 2);
	    if (ch > *hp)
	      *hp = ch;
	  }
	  break;
	case SOPT_X_HLTL:
	  {
	    /* If there is room on the page for the full cell width,
	       then it is expanded to that width.  Otherwise, it takes
	       whatever horizontal space is available.  Does not
	       change vertical size.  (This type is not useful for
	       subtables because it has the maximum table width
	       hardcoded as the page width.) */
	    int cw, ch;
	    int i, tw;

	    for (i = tw = 0; i <= 2 * t->nc; i++)
	      tw += w[i];

	    get_cell_size (d, *c, &cw, &ch, 1);
	    if (tw > d->width)
	      {
#if 0
		/* This must be wrong? */
		*wp = cw;
#else
		*wp = max (*wp, cw);
#endif
	      }
	    else if (tw - *wp + cw <= d->width)
	      *wp = max (*wp, cw);
	    else
	      {
		int proposed = d->width - (tw - *wp);
		if (proposed > *wp)
		  *wp = proposed;
	      }
	  }
	  break;
	case SOPT_X_BOTH:
	  {
	    /* First tries to expand the width of the column in order
	       to fit all the columns on the page.  But if expanding
	       the column would make the table wider than a page, then
	       it expands the height of the row instead.  Always
	       ensures that the row has enough height for at least one
	       line of text. (This type is not useful for subtables
	       because it has the maximum table width hardcoded as the
	       page width.)

	       Note the fine distinction from SOPT_X_SHSP below. */
	    int cw, ch;
	    int i, tw;

	    for (i = tw = 0; i <= 2 * t->nc; i++)
	      tw += w[i];
	    ch = *hp;
	    if (ch < som->font_height)
	      ch = som->font_height;

	    get_cell_size (d, *c, &cw, &ch, 1);
	    if (tw <= d->width && tw + cw - *wp >= d->width)
	      {
		cw = *wp;
		get_cell_size (d, *c, &cw, &ch, 2);
	      }
	    if (ch > *hp)
	      *hp = ch;
	    if (cw > *wp)
	      *wp = cw;
	  }
	  break;
	case SOPT_X_SHSP:
	  {
	    /* Tries to expand the width of the column in order to fit
	       all the columns on the page.  But if the table with the
	       expanded column would take up more than the width of a
	       page, then it expands the height of the row instead.
	       Always ensures that the row has enough height for at
	       least one line of text.  (This type is not useful for
	       subtables because it has the maximum table width
	       hardcoded as the page width.)

	       The distinction from SOPT_X_BOTH is that SOPT_X_BOTH
	       will expand the column if the table is already wider
	       than a page, assuming it's a loss to save horizontal
	       space by that time; however SOPT_X_SHSP always
	       preserves horizontal space, even if the table is
	       already more than a page in width. */

	    int cw, ch;
	    int i, tw;

	    for (i = tw = 0; i <= 2 * t->nc; i++)
	      tw += w[i];
	    ch = *hp;
	    if (ch < som->font_height)
	      ch = som->font_height;

	    get_cell_size (d, *c, &cw, &ch, 1);
	    if (tw + cw - *wp > d->width)
	      {
		cw = *wp;
		get_cell_size (d, *c, &cw, &ch, 2);
	      }
	    if (ch > *hp)
	      *hp = ch;
	    if (cw > *wp)
	      *wp = cw;
	  }
	  break;
	default:
	  assert (0);
	}
}

#if GLOBAL_DEBUGGING
/* Verifies that every cell in a table has been filled in. */
static void
check_table (som_table * t)
{
  int x, y;

  for (x = 0; x < t->nc; x++)
    for (y = 0; y < t->nr; y++)
      if (t->c[t->nc * y + x] == NULL)
	{
	  printf (_("Table cell (%d,%d) is NULL.\n"), x, y);
	  t->c[t->nc * y + x] = arena_alloc (&curtab_arena,
					     sizeof (som_empty_cell));
	  t->c[t->nc * y + x]->any.opt = SOPT_NONE;
	  t->c[t->nc * y + x]->any.con = SCON_EMPTY;
	}
}
#endif

#if 0
/* som_get_table_size() extract. */
case SOPT_X_JOIN:
{
  /* Number of cells joined. */
  int jc, jr;

  /* Joined cell metrics. */
  int cw, ch;

  /* Combined metrics of all the rows/columns that are joined. */
  int tw, th;

  /* Iterators. */
  int x2, y2;

  get_join_counts (*c, &jc, &jr);
  get_cell_size (d, *c, &cw, &ch, 0);
  for (x2 = tw = 0; x2 < jc; x2++)
    tw += w[x2 + x];
  for (y2 = th = 0; y2 < jr; y2++)
    th += h[y2 + y];
  if (cw > tw)
    {
      /* Lerp along the x dimension to evenly distribute the
         needed extra space. */
      int quot, rem, error;
      quot = (tw - cw) / jc;
      rem = (tw - cw) % jc;
      error = 0;
      for (x2 = 0; x2 < jc; x2++)
	{
	  int amt = quot;
	  error += rem;
	  if (error >= rem)
	    {
	      amt++;
	      error -= quot;
	    }
	  w[x2++] += amt;
	}
    }
  if (ch > th)
    {
      /* Lerp along the y dimension to evenly distribute the
         needed extra space. */
      int quot, rem, error;
      quot = (th - ch) / jr;
      rem = (th - ch) % jr;
      error = 0;
      for (y2 = 0; y2 < jr; y2++)
	{
	  int amt = quot;
	  error += rem;
	  if (error >= rem)
	    {
	      amt++;
	      error -= quot;
	    }
	  h[y2++] += amt;
	}
    }
}
break;
#endif
